namespace eval ten::json {

  namespace eval deparse {

    variable quotes [
      list "\"" "\\\"" / \\/ \\ \\\\ \b \\b \f \\f \n \\n \r \\r \t \\t
    ]

    variable indent
    variable indentBy 0
    variable indentIncr
    variable nl

    proc indent_one {} {
      variable indentIncr
      variable indentBy [ expr $indentBy + $indentIncr ]
      variable indent [ string repeat " " $indentBy ]
    }

    proc outdent_one {} {
      variable indentIncr
      variable indentBy [ expr $indentBy - $indentIncr ]
      variable indent [ string repeat " " $indentBy ]
    }

    proc str {str} {
      variable quotes
      return \"[ string map $quotes $str ]\"
    }

    proc num {num} {
      return $num
    }

    proc list {args} {
      variable indent
      variable nl
      set out \[
      indent_one
      foreach el [lrange $args 0 end] {
        append out $nl$indent
        append out [ deparse $el ],
      }
      outdent_one
      append out $nl$indent\]
      return $out
    }

    proc obj {args} {
      variable indent
      variable nl
      set out \{
      indent_one
      dict for {k v} $args {
        append out $nl$indent[ str $k ]:\ [ deparse $v ],
      }
      outdent_one
      append out $nl$indent\}
    }

    proc deparse {data} {
      switch -regexp [lindex $data 0] {
        ^true|false|null$ { uplevel 1 return [lindex $data 0] }
        ^num|str|obj|list$ {}
        default { error [ concat "Invalid JSON type " [lindex $data 0 0] ] }
      }
      return [ eval $data ]
    }
  }

  proc deparse_json {data {indentIncr 2}} {
    set deparse::indentBy 0
    set deparse::indentIncr $indentIncr
    if [expr $indentIncr == 0] {
      set deparse::nl ""
    } else {
      set deparse::nl "\n"
    }
    deparse::deparse $data
  }

  namespace eval tclify {

    proc str {str} { return $str }

    proc num {num} { return $num }

    proc obj {args} {
      return $args
    }

    proc tclify {data} {
      switch -regexp [lindex $data 0] {
        ^true|false|null$ { uplevel 1 return [lindex $data 0] }
        ^num|str|obj|list$ {}
        default { error [ concat "Invalid JSON type " [lindex $data 0 0] ] }
      }
      return [ eval $data ]
    }
  }

  proc tclify_json {data} {
    tclify::tclify $data
  }
}

set ex_json { list {str foo} {num 0} {obj __remote_object__ {str 512}} {null} }

puts [ ten::json::deparse_json $ex_json 2 ]

dict for {k v} [ ten::json::tclify_json [
  lindex [ ten::json::tclify_json $ex_json ] 2
] ] { puts "$k: $v" }
